/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

package uk.co.petertribble.jangle;

import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;
import java.awt.BorderLayout;
import java.awt.event.*;
import javax.swing.event.*;
import uk.co.petertribble.jingle.JingleTextPane;
import snmp.*;
import org.jfree.chart.ChartPanel;
import org.jdesktop.swingworker.SwingWorker;

/**
 * A panel displaying the snmp statistics that are available as a list or
 * tree in a left panel, and details and possibly a graphical chart in the
 * right panel.
 */
public class SnmpTreePanel extends JPanel implements TreeSelectionListener,
    ListSelectionListener, ActionListener
{

    private SnmpResources snr;
    private JingleTextPane tp;
    private JingleTextPane tpdesc;
    private JingleTextPane tpsiblings;
    private JPanel jp1;
    private JPanel jp2;
    private SnmpChart chart;

    private Exception savederror;

    private JList slist;
    private JProgressBar jpb;
    private DefaultListModel model;
    private JTabbedPane jtpl;
    private JTabbedPane jtpr;
    private static final int TAB0 = 0;
    private static final int TAB1 = 1;
    private static final int TAB2 = 2;

    private String startOID;
    private SnmpObject currentOID;
    private SnmpController sc;
    private SnmpList snl;
    private SnmpMibManager smm;
    private List <SnmpObject> oidList;

    private javax.swing.Timer timer;
    private int interval = 30;

    /**
     * Create a new SnmpTreePanel, starting exploration at the default OID
     * of 1.3.6.1.2.1.
     */
    public SnmpTreePanel() {
	this("1.3.6.1.2.1");
    }

    /**
     * Create a new SnmpTreePanel, starting exploration at the specified OID.
     *
     * @param startOID the OID to start from
     */
    public SnmpTreePanel(String startOID) {
	this.startOID = startOID;
	snr = SnmpResources.getInstance();
	setLayout(new BorderLayout());

	jtpl = new JTabbedPane();
	model = new DefaultListModel();
	oidList = new ArrayList <SnmpObject> ();
	slist = new JList(model);
	slist.addListSelectionListener(this);
	jtpl.add(snr.getString("SNMP.LIST.TEXT"), new JScrollPane(slist));

	jtpr = new JTabbedPane();

	jp1 = new JPanel(new BorderLayout());
	tp = new JingleTextPane();
	jp1.add(new JScrollPane(tp), BorderLayout.NORTH);
	// jp2 holds the chart
	jp2 = new JPanel(new BorderLayout());
	jp1.add(jp2);

	// explicit placement so the tabs can be manipulated later
	jtpr.insertTab(snr.getString("SNMP.DETAILS.TEXT"), (Icon) null, jp1,
		(String) null, TAB0);

	tpdesc = new JingleTextPane("text/plain");
	jtpr.insertTab(snr.getString("SNMP.ABOUT.TEXT"), (Icon) null,
		new JScrollPane(tpdesc), (String) null, TAB1);

	tpsiblings = new JingleTextPane("text/plain");
	jtpr.insertTab(snr.getString("SNMP.SIBLINGS.TEXT"), (Icon) null,
		new JScrollPane(tpsiblings), (String) null, TAB2);

	// split pane to hold the lot
	JSplitPane psplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
					   jtpl, jtpr);
	psplit.setOneTouchExpandable(true);
	psplit.setDividerLocation(150);
	add(psplit);
    }

    /**
     * Start exploring the snmp tree.
     *
     * @param params details of the snmp server to query
     */
    public void exploreServer(SnmpParams params) {
	sc = new SnmpController(params);
	snl = new SnmpList(sc);
	jpb = new JProgressBar(0, 1);
	jpb.setValue(0);
	jpb.setIndeterminate(true);
	jp2.add(jpb, BorderLayout.SOUTH);
	jp2.validate();
	// load data in background
	(new ExploreWorker()).execute();
    }

    /*
     * This is where the tree is walked and the data loaded. We do the snmp
     * tree walk in the background and update the gui when we're done - that
     * means building the model for the list and finishing off the progress
     * bar.
     */
    class ExploreWorker extends SwingWorker <String, Object> {
	@Override
	public String doInBackground() {
	    buildData();
	    return "done";
	}

	@Override
	protected void done() {
	    try {
		jpb.setIndeterminate(false);
		jpb.setValue(1);
		if (savederror == null) {
		    for (SnmpObject oid : oidList) {
			model.addElement(oid);
		    }
		    buildTree();
		    startLoop();
		} else {
		    showError();
		}
		jp2.removeAll();
		jp2.validate();
	    } catch (Exception e) {}
	}
    }

    /**
     * Start the timer, so the display will continually update.
     */
    public void startLoop() {
	if (timer == null) {
	    timer = new javax.swing.Timer(interval*1000, this);
	}
	timer.start();
    }

    private void setPanel(DefaultMutableTreeNode node) {
	if (node != null) {
	    if (node.getUserObject() instanceof SnmpObject) {
		showObject((SnmpObject) node.getUserObject());
	    } else if (node.getUserObject() instanceof String) {
		showObject((String) node.getUserObject());
	    }
	}
    }

    /*
     * Show the MIB details for a fake parent node.
     * We only do any work if it exists - if we can't find a MIB entry
     * then we leave the display unchanged.
     */
    private void showObject(String oid) {
	if (smm == null) {
	    smm = SnmpMibManager.getInstance();
	}
	String s = smm.getNearestMibForOID(oid);
	if (s != null) {
	    if (chart != null) {
		chart.stopLoop();
		jp2.removeAll();
	    }
	    tpdesc.setText(s);
	    jtpr.setSelectedIndex(TAB1);
	    jtpr.setEnabledAt(TAB0, false);
	    jtpr.setEnabledAt(TAB2, false);
	}
    }

    private void showObject(SnmpObject sno) {
	if (smm == null) {
	    smm = SnmpMibManager.getInstance();
	}
	jtpr.setEnabledAt(TAB0, true);
	jtpr.setEnabledAt(TAB2, true);
	// stop any current chart
	if (chart != null) {
	    chart.stopLoop();
	    jp2.removeAll();
	}
	currentOID = sno;
	setText(sno);
	String oid = sno.toString();
	// show a chart if it looks like a number
	String typeString = sno.getTypeString();
	if (!typeString.equals("OctetString")
	    && !typeString.equals("IPAddress")
	    && !typeString.equals("Null")
	    && !typeString.equals("ObjectIdentifier")) {
	    chart = new SnmpChart(sc, oid, smm.prettifyOID(oid));
	    jp2.add(new ChartPanel(chart.getChart()));
	}
	jp2.validate();
	jp1.validate();
	String s = smm.getNearestMibForOID(oid);
	tpdesc.setText((s == null) ? "" : s);
	showSiblings(sno);
    }

    private void showSiblings(SnmpObject sno) {
	StringBuilder sb = new StringBuilder();
	for (SnmpObject so : snl.getSiblings(oidList, sno.toString())) {
	    sb.append(smm.prettifyOID(so.toString())).append("=");
	    sb.append(so.valueString()).append("\n");
	}
	snl.getCousins(oidList, sno.toString());
	tpsiblings.setText(sb.toString());
    }

    private void setText(SnmpObject sno) {
	StringBuilder sb = new StringBuilder();
	sb.append("OID: ").append(smm.prettifyOID(sno.toString())).append("\n");
	String typeString = sno.getTypeString();
	sb.append("Type: ").append(typeString).append("\n");
	sb.append("Value: ");
	sb.append(sno.valueString());
	if (typeString.equals("OctetString")) {
	    sb.append(" (hex: ");
	    sb.append(((SNMPOctetString) sno.getValue()).toHexString());
	    sb.append(")");
	}
	if (typeString.equals("ObjectIdentifier")) {
	    sb.append(" = ");
	    sb.append(smm.prettifyOID(sno.getValue().toString()));
	}
	sb.append("\n");
	tp.setText(sb.toString());
    }

    /**
     * Set the update interval.
     *
     * @param i the desired update delay in seconds
     */
    public void setDelay(int i) {
	interval = i;
	if (timer != null) {
	    timer.setDelay(i*1000);
	}
	if (chart != null) {
	    chart.setDelay(i);
	}
    }

    private void updateCurrent() {
	if (currentOID != null) {
	    updateOID(currentOID);
	    setText(currentOID);
	}
    }

    private void updateOID(SnmpObject sno) {
	try {
	    sno.setValue(sc.getValue(sno.toString()).getValue());
	} catch (SnmpException sne) {
	    // FIXME do summat
	}
    }

    // retrieve the data via an snmp walk of the tree
    private void buildData() {
	try {
	    oidList = snl.getList(startOID);
	} catch (SnmpException sne) {
	    savederror = sne;
	}
    }

    private void showError() {
	if (savederror != null) {
	    System.out.println(savederror.getMessage());
	    JOptionPane.showMessageDialog(this,
					savederror.getMessage(),
					"SNMP error",
					JOptionPane.ERROR_MESSAGE);
	    savederror = null;
	}
    }

    private void buildTree() {
	/*
	 * We create a DefaultMutableTreeNode for each element in the list
	 * and add it to a Hash with the name as the key. Then we walk the
	 * list - if a node has a parent, we add it to the parent, else we
	 * need to create intermediate nodes to fill in the tree.
	 */
	// root node
	DefaultMutableTreeNode rootNode = new DefaultMutableTreeNode("SNMP");
	// real nodes
	Map <String, DefaultMutableTreeNode> m =
			new HashMap <String, DefaultMutableTreeNode> ();
	// intermediate nodes
	Map <String, DefaultMutableTreeNode> m2 =
			new HashMap <String, DefaultMutableTreeNode> ();
	for (SnmpObject sno : oidList) {
	    m.put(sno.toString(), new DefaultMutableTreeNode(sno));
	}
	for (SnmpObject sno : oidList) {
	    String s = sno.toString();
	    DefaultMutableTreeNode mynode = m.get(s);
	    String parent = getParentOID(s);
	    while (parent != null) {
		if (m.containsKey(parent)) {
		    /*
		     * If there's a parent node, add this as a child.
		     * This almost never happens.
		     */
		    m.get(parent).add(mynode);
		    mynode = m.get(parent);
		} else if (m2.containsKey(parent)) {
		    /*
		     * If we've already created a parent node, add this node
		     * to that.
		     */
		    m2.get(parent).add(mynode);
		    mynode = m2.get(parent);
		} else {
		    /*
		     * Need to create a new node
		     */
		    DefaultMutableTreeNode pnode =
			new DefaultMutableTreeNode(parent);
		    pnode.add(mynode);
		    m2.put(parent, pnode);
		    mynode = pnode;
		}
		parent = getParentOID(parent);
	    }
	    rootNode.add(mynode);
	}
	JTree stree = new JTree(rootNode);
	stree.addTreeSelectionListener(this);
	jtpl.add(snr.getString("SNMP.TREE.TEXT"), new JScrollPane(stree));
    }

    // duplicated from SnmpMibManager
    private String getParentOID(String s) {
	int i = s.lastIndexOf('.');
	return (i < 0) ? null: s.substring(0, i);
    }

    // handle ListSelectionListener events
    public void valueChanged(ListSelectionEvent e) {
        if (!e.getValueIsAdjusting() && (slist.getSelectedIndex() != -1)) {
	    Object o = slist.getSelectedValue();
	    if (o instanceof SnmpObject) {
		showObject((SnmpObject) o);
	    }
        }
    }

    // handle TreeSelectionListener events
    public void valueChanged(TreeSelectionEvent e) {
	TreePath tpth = e.getNewLeadSelectionPath();
	if (tpth != null) {
	    setPanel((DefaultMutableTreeNode) tpth.getLastPathComponent());
	}
    }

    // handle timer events
    public void actionPerformed(ActionEvent e) {
	updateCurrent();
    }
}
